/*
 * milter-spamc.c
 *
 * Copyright 2003, 2004 by Anthony Howe. All rights reserved.
 *
 * The following should be added to the sendmail.mc file:
 *
 *	INPUT_MAIL_FILTER(
 *		`milter-spamc',
 *		`S=unix:/var/lib/milter-spamc/socket, T=S:30s;R:3m'
 *	)dnl
 *	define(
 *		`confMILTER_MACROS_CONNECT', confMILTER_MACROS_CONNECT`,
 *		{client_addr}, {client_name}, {client_port}, {client_resolve}'
 *	)dnl
 */

/***********************************************************************
 *** Leave this header alone. Its generate from the configure script.
 ***********************************************************************/

#include "milter-spamc.h"

/***********************************************************************
 *** You can change the stuff below if the configure script doesn't work.
 ***********************************************************************/

/*
 * Define NDEBUG for production, undefine for testing (especially
 * on a system with an existing SpamAssassin milter).
 */
#define NDEBUG

#ifndef RUN_AS_USER
#define RUN_AS_USER			"smmsp"
#endif

#ifndef RUN_AS_GROUP
#define RUN_AS_GROUP			"smmsp"
#endif

#ifndef DEFAULT_SPAMD_HOST
#define DEFAULT_SPAMD_HOST		"127.0.0.1"
#endif

#ifndef DEFAULT_SPAMD_PORT
#define DEFAULT_SPAMD_PORT		783
#endif

#ifndef DEFAULT_SOCKET_TIMEOUT
#define DEFAULT_SOCKET_TIMEOUT		120
#endif

#ifndef DEFAULT_HEADER_BUFFER_SIZE
#define DEFAULT_HEADER_BUFFER_SIZE	8192
#endif

#ifndef DEFAULT_SPAM_SUBJECT_TAG
#define DEFAULT_SPAM_SUBJECT_TAG	"[SPAM]"
#endif

#ifndef DEFAULT_SPAM_LEVEL_CHARACTER
#define DEFAULT_SPAM_LEVEL_CHARACTER	'x'
#endif

/*
 * Choose between CHECK, REPORT_IFSPAM, and REPORT.  
 * PROCESS and SYMBOLS not supported.
 */
#ifndef DEFAULT_SPAMC_COMMAND
#define DEFAULT_SPAMC_COMMAND		"REPORT"
#endif

#ifndef STATEDIR
#define STATEDIR			"/var/lib/milter-spamc"
#endif

/*
 * For when the mutex is a file lock.
 */
#ifndef THE_MUTEX
#define THE_MUTEX			STATEDIR "/mutex"
#endif

/*
 * Allow time for smfi_stop() to cleanly shutdown the milter and exit
 * after a critical error, otherwise we fallback on abort().
 */
#ifndef DELAY_BEFORE_ABORT
#define DELAY_BEFORE_ABORT		60
#endif

#define CLOSE_STANDARD_HANDLES

/***********************************************************************
 *** No configuration below this point.
 ***********************************************************************/

/* Re-assert this macro just in case. May cause a compiler warning. */
#define _REENTRANT	1

#include <sys/types.h>
#include <pwd.h>
#include <grp.h>
#include <ctype.h>
#include <errno.h>
#include <time.h>
#include <signal.h>
#include <stdarg.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <syslog.h>
#include <unistd.h>
#include <sys/stat.h>
#include <netdb.h>

#include <libmilter/mfapi.h>

#include <com/snert/lib/version.h>
#include <com/snert/lib/io/Socket.h>
#include <com/snert/lib/sys/Mutex.h>
#include <com/snert/lib/mail/parsePath.h>
#include <com/snert/lib/mail/smdb.h>
#include <com/snert/lib/util/Buf.h>
#include <com/snert/lib/util/Text.h>
#include <com/snert/lib/util/Vector.h>
#include <com/snert/lib/util/getopt.h>

#if LIBSNERT_MAJOR < 1 || LIBSNERT_MINOR < 31
# error "LibSnert/1.31 or better is required"
#endif

#if defined(LIBSNERT_WITH_BERKELEY_DB) && defined(MILTER_WITH_BERKELEY_DB)
# define HAVE_BERKELEY_DB
#else
# undef HAVE_BERKELEY_DB
#endif

/***********************************************************************
 *** Global Variables & Constants
 ***********************************************************************/

#ifndef MILTER_CHUNK_SIZE
#define MILTER_CHUNK_SIZE	65535
#endif

/* See RFC 2821 section 4.5.3.1 Size limits and minimums */
#define SMTP_LINE_SIZE		1000
#define SMTP_REPLY_SIZE		512
#define IPV6_LENGTH		((128/16)*4+(128/16)-1)

#define DOMAIN_LENGTH		255
#define LOCAL_PART_LENGTH	128
#define EMAIL_LENGTH		(LOCAL_PART_LENGTH+1+DOMAIN_LENGTH)

/* NOTE that the LOCAL_PART_LENGTH in RFC 2821 is limited to 64
 * octets, but some stupid mailing list management software use 
 * Variable Envelope Return Paths (VERP) or the Sender Rewriting
 * Scheme (SRS), in order to encode another address within the
 * local-part of an address. If the address being encoded is 
 * already EMAIL_LENGTH octets in length, then the new encoded
 * address will NOT be conformant with RFC 2821 size limits for
 * the local-part.
 *
 * The LOCAL_PART_LENGTH here is doubled as a partial concesion
 * to these popular yet broken schemes. It appears to work for
 * majority of cases.
 */

#define PARSE_PATH		"<%384[^>]>%n"		/* EMAIL_LENGTH = 384 */
#define PARSE_MAILBOX		"%128[^@]%n@%255s%n"	/* LOCAL_PART_LENGTH & DOMAIN_LENGTH */
#define PARSE_PLUS_DETAIL	"%128[^+]+%128[^@]"	/* LOCAL_PART_LENGTH = 128 */
#define FORMAT_TIMESTAMP	"; %a, %d %b %Y %H:%M:%S %z"

#define MX_INFO_FORMAT		"MX %d '%s' [%s] for <%s>"
#define MX_INFO_ARGS		mx->preference, mx->name, data->ipv6, data->mailAddress

#define SMTP_OK			250
#define SMTP_IS_OK(x)		(200 <= (x) && (x) < 300)
#define SMTP_IS_DEFER(x)	(300 <= (x) && (x) < 400)
#define SMTP_IS_TEMP(x)		(400 <= (x) && (x) < 500)
#define SMTP_IS_PERM(x)		(500 <= (x) && (x) < 600)
#define SMTP_IS_VALID(x)	(200 <= (x) && (x) < 600)

typedef struct {
	int isLocal;				/* rcpt_mailer == 'local or 'cyrusv2' */
	long length;				/* address length */
	long capacity;				/* address buffer capacity */
	char *address;				/* recipient address with < and > */
} *Rcpt;

typedef struct {
	SMFICTX *ctx;				/* pointer to this */
	Socket server;				/* per message */
	ParsePath *mail;			/* per message */
	Vector rcpts;				/* per message */
	Buf headers;				/* per message */
	time_t now;				/* per message */
	int byPass;				/* per message */
	int isSent;				/* per message */
	int isWhiteIP;				/* per connection */
	int localRcptCount;			/* per message */
	int untrustedRcptCount;			/* per message */
	int hasSubject;				/* per message */
	int hasSpamFlag;			/* per message */
	int hasSpamLevel;			/* per message */
	int hasSpamReport;			/* per message */
	int hasSpamStatus;			/* per message */
	int precedence;				/* per message */
	long indexReceived;			/* per message */
	char *qid;				/* per message */
	char *localRcpt;			/* per message */
	char helo[DOMAIN_LENGTH+1];		/* per connection */
	char line[SMTP_LINE_SIZE+1];		/* general purpose */
	char subject[SMTP_LINE_SIZE+1];		/* subject header */
	char replyLine[SMTP_REPLY_SIZE+1];	/* setReply(), vSetReply() */
} *workspace;

#define PRECEDENCE_SPECIAL_DELIVERY	4
#define PRECEDENCE_FIRST_CLASS		3
#define PRECEDENCE_LIST			2
#define PRECEDENCE_JUNK			1
#define PRECEDENCE_BULK			0

static char *precedence[] = {
	"bulk", "junk", "list", "first-class", "special-delivery", (char *) 0
};

#define DEBUG_ALL		(~0L)
#define DEBUG_INFO		1
#define DEBUG_TRACE		2
#define DEBUG_SPAMD		4
#define DEBUG_DATABASE		8
#define DEBUG_SOCKETS		16
#define DEBUG_DATABASE_LOG	128

#define COPY_TO			0
#define REDIRECT_TO		1
#define REDIRECT_TO_FORMAT	2

static long debug;
static long socketTimeout;

static char *spamcCommand;
static int alwaysAddReport;
static int forceSpamdUser;
static int redirectSpam;
static int spamdPort;
static int untrustedRcptLimit;
static int appendScoreToSubjectTag;
static char *spamdHost;
static char *spamdUser;
static char *copyAllTo;
static char *copyHamTo;
static char *copySpamTo;
static char *rcptReformat;
static char *subjectTag;
static char *subjectScoreTag;
static char *sendmailCf;
static char *milterSocket;
static long rejectExcessScore;
static long discardExcessScore;

static char hostaddr[IPV6_LENGTH+1];
static char hostname[DOMAIN_LENGTH+1];

#ifdef SMFIS_REJECT_AND_DELIVER
static long rejectAndDeliverExcessScore;
#endif

static char noqueue[] = "NOQUEUE";

static char getoptString[] = "ACSa:b:B:f:F:c:d:D:h:l:r:R:s:t:u:U:v:H:P:?";

static char usageMessage[] =
"\n"
"usage:\n"
"\n"
"milter-spamc [-A|-C][-a email][-b email|-B email][-f cf][-F format]\n"
"      [-c email][-d excess][-l limit][-r excess][-R excess][-s tag]\n"
"      [-S][-t timeout][-u user|-U user][-v debug][-H host][-P port]\n"
"      socket\n"
"\n"
"-a email\tcopy all mail to email address\n"
"-A\t\talways add the X-Spam-Report: header report of rules triggered\n"
"-b email\tcopy junk mail to email address\n"
"-B email\tredirect junk mail to email address\n"
"-c email\tcopy clean mail to email address\n"
"-C\t\tcheck for spam only, disable X-Spam-Report: header completely\n"
"-d excess\tdiscard spam messages that exceed the threshold by N\n"
"-f cf\t\tpathname of sendmail.cf file for access database whitelists\n"
"-F format\tredirect junk mail to reformated recipient addresses\n"
"-h\t\tthis usage summary\n"
"-H host\t\tspamd host name or IP\n"
"-l limit\tlimit number of untrusted recipients per message\n"
"-P port\t\tspamd port address, -1 to disable spamd filtering\n"
"-r excess\treject spam that exceeds the threshold by N, overrides delivery\n"
"-R excess\treject spam that exceeds the threshold by N, allow delivery\n"
"-s tag\t\tspam subject tag\n"
"-S\t\tappend the spam score to the spam subject tag\n"
"-t timeout\tspamd I/O timeout in seconds, 0 indefinite\n"
"-u user\t\tdefault user account to process messages with\n"
"-U user\t\tforce message processing with this user account\n"
"-v debug\tthe debug bit mask\n"
"\n"
MILTER_VERSION " " MILTER_COPYRIGHT "\n"
"\n"
;

#ifdef NDEBUG
/* Headers as they should appear when in production. */

#  define X_SCANNED_BY			"X-Scanned-By"
#  define X_SPAM_FLAG			"X-Spam-Flag"
#  define X_SPAM_LEVEL			"X-Spam-Level"
#  define X_SPAM_REPORT			"X-Spam-Report"
#  define X_SPAM_STATUS			"X-Spam-Status"
#  define X_ORIGINAL_RECIPIENT		"X-Original-Recipient"

#else

/* Headers as they should appear when testing, especially in
 * conjunction with other SpamAssassin milters that handle
 * the same set of headers. This allows me to test and see
 * the headers added by my milter and compare with those of
 * others.
 */

#  define X_SCANNED_BY			"X-" MILTER_NAME "-Scanned-By"
#  define X_SPAM_FLAG			"X-" MILTER_NAME "-Flag"
#  define X_SPAM_LEVEL			"X-" MILTER_NAME "-Level"
#  define X_SPAM_REPORT			"X-" MILTER_NAME "-Report"
#  define X_SPAM_STATUS			"X-" MILTER_NAME "-Status"
#  define X_ORIGINAL_RECIPIENT		"X-" MILTER_NAME "-Original-Recipient"

#  define DEFAULT_SPAM_SUBJECT_TAG	"[" MILTER_NAME "]"

#endif

/***********************************************************************
 *** Support routines.
 ***********************************************************************/

void
die()
{
	smfi_stop();
	sleep(DELAY_BEFORE_ABORT);

	/* We should not normally reach this if the smfi_stop()
	 * kills this thread and the milter in a reasonable
	 * amount of time.
	 */
	abort();
}

#if defined(HAVE_BERKELEY_DB)
static void *theMutex;

void
lockTheMutex(void)
{
	if (MutexLock(theMutex)) {
		syslog(LOG_CRIT, "stopping milter, failed to lock mutex");
		die();
	}
}

void
unlockTheMutex(void)
{
	if (MutexUnlock(theMutex)) {
		syslog(LOG_CRIT, "stopping milter, failed to unlock mutex");
		die();
	}
}
#endif

sfsistat
nullWorkspaceError(SMFICTX *ctx)
{
	/* Without the workspace we can't do a hell of a lot. */
#ifdef NULL_WORKSPACE_ERROR_MESSSAGE
/* smfi_setreply() alloocates memory and if we arrived here, we probably
 * have a memory space problem and cannot rely on this to work correctly.
 */
	(void) smfi_setreply(ctx, "421", "4.3.2", "close connection");
#endif
	syslog(LOG_ERR, "internal error: null workspace");
	return SMFIS_TEMPFAIL;
}

sfsistat
vSetReply(workspace data, int code, const char *ecode, const char *fmt, va_list args)
{
	sfsistat rc;
	char rcode[4], xcode[6];		

	if (SMTP_IS_PERM(code))
		rc = SMFIS_REJECT;
	else if (SMTP_IS_TEMP(code))
		rc = SMFIS_TEMPFAIL;
	else
		rc = SMFIS_CONTINUE;

	/* Make sure the return code is present and in an acceptable range. */
	if (!SMTP_IS_VALID(code))
		code = 550;

	/* Convert int value to 3-digit string for smfi_setreply(). */
	rcode[3] = '\0';
	rcode[2] = '0' + (code / 1   % 10);
	rcode[1] = '0' + (code / 10  % 10);
	rcode[0] = '0' + (code / 100 % 10);

	if (ecode == (const char *) 0) {
		if (isdigit(fmt[0])) {
			ecode = strncpy(xcode, fmt, sizeof(xcode)-1);
			xcode[sizeof(xcode)-1] = '\0';
			fmt += sizeof(xcode);
		} else {
			ecode = strncpy(xcode, "x.7.1", sizeof xcode);
			xcode[0] = rcode[0];
		}
	}

	/* Build reply message. */
	(void) vsnprintf(data->replyLine, SMTP_LINE_SIZE, fmt, args);

	if (debug & DEBUG_TRACE)
		syslog(LOG_DEBUG, "%s: next reply %s %s %s", data->qid, rcode, ecode, data->replyLine);

	/* Tell sendmail what error to report to the sender. */
	(void) smfi_setreply(data->ctx, rcode, (char *) ecode, data->replyLine);

	return rc;
}

sfsistat
setReply(workspace data, int rcode, const char *xcode, const char *fmt, ...)
{
	sfsistat rc;
	va_list args;
	
	va_start(args, fmt);
	rc = vSetReply(data, rcode, xcode, fmt, args);
	va_end(args);

	return rc;
}

/*
 * Write buffer, logging only errors.
 */
int
writebuffer(workspace data, char *buf, long size)
{
	int error;
	int rc = 0;
	long nbytes, length;

	/* Do not block on sending to the spamd server just yet. */
	(void) SocketSetNonBlocking(data->server, 1);

	for (length = 0; length < size; length += nbytes) {
		if ((nbytes = SocketWrite(data->server, buf + length, size - length)) < 0) {
			error = SocketGetError(data->server);
			syslog(LOG_ERR, "%s: SocketWrite() error: %s (%d)", data->qid, strerror(error), error);
			goto error0;
		}

		/* Now wait for the output to be sent to the spamd server. */
		if (SocketWaitForOutput(data->server, socketTimeout) != 0) {
			syslog(LOG_ERR, "%s: timeout before output sent to SPAMD server", data->qid);
			goto error0;
		}
	}

	rc = 1;
error0:
	(void) SocketSetNonBlocking(data->server, 0);

	return rc;
}

/*
 * Log line to be sent then write line.
 */
int
writeline(workspace data, char *line, long size)
{
	if (debug & DEBUG_SPAMD)
		syslog(LOG_DEBUG, "%s: > %s", data->qid, line);

	return writebuffer(data, line, size < 0 ? strlen(line) : size);
}

/*
 * Log amount of data sent then write data.
 */
int
writechunk(workspace data, char *chunk, long size)
{
	if (debug & DEBUG_SPAMD)
		syslog(LOG_DEBUG, "%s: > [%ld bytes of data]", data->qid, size);

	return writebuffer(data, chunk, size);
}

int
readline(workspace data, char *line, int linesize)
{
	long length;
	int rc, error;

	rc = -1;

	if (linesize < sizeof (long)) {
		syslog(LOG_CRIT, "%s: line buffer too small (%d)", data->qid, linesize);
		die();
	}

	error = SocketWaitForInput(data->server, socketTimeout);

	if (error == ETIMEDOUT) {
		syslog(LOG_ERR, "%s: timeout before input from SPAMD server", data->qid);
		goto error0;
	}

	if (error != 0) {
		syslog(LOG_ERR, "%s: input error from SPAMD server: %s (%d)", data->qid, strerror(error), error);
		goto error0;
	}

	length = SocketReadLine(data->server, line, linesize);

	if ((error = SocketGetError(data->server)) != 0 || length == -1) {
		syslog(LOG_ERR, "%s: read error from SPAMD server: %s (%d)", data->qid, strerror(error), error);
		goto error0;
	}

	if (debug & DEBUG_SPAMD)
		syslog(LOG_DEBUG, "%s: < %s", data->qid, line);

	rc = length;
error0:
	return rc;
}

/***********************************************************************
 *** Handlers
 ***********************************************************************/

/*
 * Open and allocate per-connection resources.
 */
sfsistat
filterOpen(SMFICTX *ctx, char *hostname, _SOCK_ADDR *hostaddr)
{
	int rc;
	workspace data;

	if (debug & DEBUG_TRACE)
		syslog(LOG_DEBUG, "create per-connection resources...");

	if ((data = calloc(1, sizeof *data)) == (workspace) 0)
		goto error0;

	if ((data->headers = BufCreate(DEFAULT_HEADER_BUFFER_SIZE)) == (Buf) 0)
		goto error1;

	if ((data->rcpts = VectorCreate(10, free)) == (Vector) 0)
		goto error2;

	if (smfi_setpriv(ctx, (void *) data) == MI_FAILURE)
		goto error3;

	/* Remember the context we belong to. Useful when passing around
	 * numerous arguements.
	 */
	data->ctx = ctx;
	data->isWhiteIP = 0;
	data->qid = noqueue;
	
#ifdef HAVE_BERKELEY_DB
{
	char *client_addr, *client_name;

	if (!MutexLock(theMutex)) {
		/* Check this once per message instead of repeatly each
		 * time we access the DB. If the database changes after
		 * this, its not to critical (I hope), that we'll catch
		 * it the next time through.
		 */
		if (smdbAliases != (Vector) 0) {
			int i;
			for (i = 0; i < VectorLength(smdbAliases); i++)
				(void) smdbReload(VectorGet(smdbAliases, i));
		}
		(void) smdbReload(smdbAccess);
		unlockTheMutex();
	}

	client_addr = smfi_getsymval(data->ctx, "{client_addr}");
	client_name = smfi_getsymval(data->ctx, "{client_name}");

	if (client_addr == (char *) 0 || client_name == (char *) 0) {
		syslog(LOG_ERR, "%s: {client_addr} and {client_name} must be available, see cf/README `confMILTER_MACROS_CONNECT'", data->qid);
		goto skip0;
	}

	/* BY-PASS
	 *
	 * Black and white lists IP handled through Sendmail's access
	 * database. The lookup order is:
	 *
	 *	Connect:a.b.c.d
	 *	Connect:a.b.c
	 *	Connect:a.b
	 *	Connect:a
	 *	a.b.c.d
	 *	a.b.c
	 *	a.b
	 *	a
	 *
	 * This Sendmail syntax limits the netmasks to /32, /24, /16, /8,
	 * which are the most common cases, but not so flexible as full
	 * range netmasks. Doing that would require a separate preloaded
	 * network/mask firewall-like list from another file since the
	 * access database syntax does not express this information well.
	 */
	if (smdbAccessIsIpOk(smdbAccess, "connect:", client_addr) == SMDB_ACCESS_OK)
		data->isWhiteIP = 1;

	/* BY-PASS
	 *
	 * If sendmail successfully resolves the client connection (PTR-A-PTR 
	 * lookup) and there is a Connect:domain OK or RELAY entry, then white
	 * list the connection. We don't reject on this since Sendmail will act
	 * on it itself before calling the milter.
	 *
	 * If the client IP resolved and matched, then the lookup order is:
	 *
	 *	Connect:some.sub.domain.tld
	 *	Connect:sub.domain.tld
	 *	Connect:domain.tld
	 *	Connect:tld	 
	 *	some.sub.domain.tld
	 *	sub.domain.tld
	 *	domain.tld
	 *	tld	 
	 *
	 * If the client IP did not resolve or match, then the lookup order is:
	 *
	 *	Connect:[ip]
	 *	[ip]
	 */
	else if (smdbAccessIsDomainOk(smdbAccess, "connect:", client_name) == SMDB_ACCESS_OK)
		data->isWhiteIP = 1;
skip0:
	;
}
#endif	
	return SMFIS_CONTINUE;
error3:
	VectorDestroy(data->rcpts);
error2:
	BufDestroy(data->headers);
error1:
	free(data);
error0:
	return SMFIS_REJECT;
}

sfsistat
filterHelo(SMFICTX * ctx, char *helohost)
{
	workspace data;

	if ((data = (workspace) smfi_getpriv(ctx)) == (workspace) 0)
		return nullWorkspaceError(ctx);

	(void) strncpy(data->helo, helohost, sizeof(data->helo));

	return SMFIS_CONTINUE;
}

sfsistat
filterMail(SMFICTX *ctx, char **args)
{
	sfsistat rc;
	workspace data;
	const char *error;
	char *auth_authen;

	if ((data = (workspace) smfi_getpriv(ctx)) == (workspace) 0)
		return nullWorkspaceError(ctx);

	/* Reset per-message variables, since this function can be
	 * called several times per message or connection, ie.
	 * MAIL-RSET-MAIL or MAIL-RCPT-DATA-MAIL sequences.
	 */

	VectorRemove(data->rcpts, 0, VectorLength(data->rcpts));
	BufSetLength(data->headers, 0);
	data->untrustedRcptCount = 0;
	data->localRcptCount = 0;

	if (data->localRcpt != spamdUser)
		free(data->localRcpt);

	data->byPass = spamdPort < 0 || data->isWhiteIP;
	data->precedence = PRECEDENCE_FIRST_CLASS;
	data->qid = smfi_getsymval(ctx, "i");
	data->localRcpt = (char *) 0;
	data->indexReceived = -1;
	data->isSent = 0;

	(void) time(&data->now);

	/* Make sure its closed before we start the next message. */
	SocketClose(data->server);
	data->server = (Socket) 0;

	/* We assume that an authenticated user is trusted not to send spam. */
	auth_authen = smfi_getsymval(data->ctx, "{auth_authen}");
	if (auth_authen != (char *) 0) {
		if (debug & (DEBUG_INFO|DEBUG_TRACE))
			syslog(LOG_INFO, "%s: authenticated \"%s\", skipping", data->qid, auth_authen);
		data->byPass = 1;
		goto skip0;
	}
	
	if (data->mail != (ParsePath *) 0)
		free(data->mail);
	
	if ((error = parsePath(args[0], 1, &data->mail)) != (const char *) 0)		
		return setReply(data, 553, (const char *) 0, error);

#ifdef HAVE_BERKELEY_DB
	if (smdbAccessGetEmailDomainUserAccess(smdbAccess, "spam:", data->mail->address.string, data->mail->domain.string, data->mail->localLeft.string) == SMDB_ACCESS_OK) {
		if (debug & (DEBUG_INFO|DEBUG_TRACE))
			syslog(LOG_INFO, "%s: sender <%s> spam whitelisted, skipping", data->qid, data->mail->address.string);
		data->byPass = 1;
		goto skip0;
	}
	
# ifdef TEST_ACCESS_DATABASE_FROM_TAG	
	if (smdbAccessIsEmailDomainUserOk(smdbAccess, "from:", data->mail->address.string, data->mail->domain.string, data->mail->localLeft.string) == SMDB_ACCESS_OK) {
		if (debug & (DEBUG_INFO|DEBUG_TRACE))
			syslog(LOG_INFO, "%s: sender <%s> whitelisted, skipping", data->qid, data->mail->address.string);
		data->byPass = 1;
		goto skip0;
	}
# endif
#endif
skip0:	
	return SMFIS_CONTINUE;
}

sfsistat
filterRcpt(SMFICTX *ctx, char **args)
{
	sfsistat rc;
	workspace data;
	ParsePath *path;
	const char *error;
	char *rcpt_addr, *rcpt_host, *rcpt_mailer;

	if ((data = (workspace) smfi_getpriv(ctx)) == (workspace) 0)
		return nullWorkspaceError(ctx);

	rcpt_addr = smfi_getsymval(data->ctx, "{rcpt_addr}");
	rcpt_host = smfi_getsymval(data->ctx, "{rcpt_host}");
	rcpt_mailer = smfi_getsymval(data->ctx, "{rcpt_mailer}");
		
	if (debug & DEBUG_TRACE)
		syslog(
			LOG_DEBUG, "%s: RCPT='%s' rcpt_addr='%s' rcpt_host='%s' rcpt_mailer='%s'",
			data->qid, args[0], rcpt_addr, rcpt_host, rcpt_mailer
		);

	if (data->byPass)
		return SMFIS_CONTINUE;

	if ((error = parsePath(args[0], 0, &path)) != (const char *) 0)		
		return setReply(data, 553, (const char *) 0, error);

#ifdef HAVE_BERKELEY_DB
	if (smdbAccessGetEmailDomainUserAccess(smdbAccess, "spam:", path->address.string, path->domain.string, path->localLeft.string) == SMDB_ACCESS_OK) {
		if (debug & (DEBUG_INFO|DEBUG_TRACE))
			syslog(LOG_INFO, "%s: recipient <%s> spam whitelisted, skipping", data->qid, path->address.string);
		data->byPass = 1;
		free(path);
		return SMFIS_CONTINUE;
	}
	
# ifdef TEST_ACCESS_DATABASE_TO_TAG	
	if (smdbAccessIsEmailDomainUserOk(smdbAccess, "to:", path->address.string, path->domain.string, path->localLeft.string) == SMDB_ACCESS_OK) {
		if (debug & (DEBUG_INFO|DEBUG_TRACE))
			syslog(LOG_INFO, "%s: recipient <%s> whitelisted, skipping", data->qid, path->address.string);
		data->byPass = 1;
		free(path);
		return SMFIS_CONTINUE;
	}
# endif
#endif

	/* TEST if the untrustedRcptCount has exceed the -l limit. */
	if (0 < untrustedRcptLimit && untrustedRcptLimit <= data->untrustedRcptCount)
		return setReply(data, 452, "4.5.3", "too many untrusted recipients");
/*		return setReply(data, 421, "4.3.2", "close connection"); */

	path->isLocal = *rcpt_host == '\0' || strncmp(rcpt_mailer, "cyrus", 5) == 0;

	/* Remember last local user. Ideally we should also attempt to
	 * expand mail aliases, but that means lots of Berkeley DB issues,
	 * so for the time being we'll ignore it and hope its not that
	 * important.
	 *
	 * Normally {rcpt_mailer} == 'local' or 'cyrusv2', {rcpt_host} == ''
	 * and {rcpt_addr} == account. However, some sites do some special
	 * rulesets for cyrus or cyrusv2 where the {rcpt_host} != '', yet
	 * the user is a local account.
	 */
	if (path->isLocal && getpwnam(rcpt_addr) != (struct passwd *) 0) {
		/* Free NULL or a previous local recipient. */
		free(data->localRcpt);

		if (debug & DEBUG_TRACE)
			syslog(LOG_DEBUG, "%s: address %s has an account", data->qid, args[0]);

		/* Assign local recipient. Don't worry if NULL, handled in filterBody(). */
		data->localRcpt = TextDup(rcpt_addr);
	}

	data->localRcptCount += path->isLocal;
	data->untrustedRcptCount++;

	if (VectorAdd(data->rcpts, path)) {
		free(path);
		return setReply(data, 452, "4.3.2", "out of memory, cannot add recipient to list");
	}

	return SMFIS_CONTINUE;
}

sfsistat
filterHeader(SMFICTX *ctx, char *name, char *value)
{
	int i;
	workspace data;

	if ((data = (workspace) smfi_getpriv(ctx)) == (workspace) 0)
		return nullWorkspaceError(ctx);

	if (debug & DEBUG_TRACE)
		syslog(LOG_DEBUG, "%s: filterHeader(%lx, '%s', '%s')", data->qid, (long) ctx, name, value);

	if (data->byPass)
		return SMFIS_CONTINUE;

	if (data->indexReceived < 0 && TextInsensitiveCompare(name, "Received") == 0) {
		data->indexReceived = BufLength(data->headers);
		if (debug & DEBUG_TRACE)
			syslog(LOG_DEBUG, "%s: indexReceived=%ld", data->qid, data->indexReceived);
	}

	if (TextInsensitiveCompare(name, "Subject") == 0) {
		(void) strncpy(data->subject, value, SMTP_LINE_SIZE - strlen(subjectTag) - 1);
		data->hasSubject = 1;
	}

	/* There appears to be a bug in libmilter. smfi_chgheader() is
	 * suppose to change or add the header if missing, but it don't
	 * work as advertised, so we have to track the precense of these
	 * headers and call smfi_chgheader() or smfi_addheader() as required.
	 */
	if (TextInsensitiveCompare(name, X_SPAM_FLAG) == 0) {
		data->hasSpamFlag = 1;
		return SMFIS_CONTINUE;
	}		
	if (TextInsensitiveCompare(name, X_SPAM_LEVEL) == 0) {
		data->hasSpamLevel = 1;
		return SMFIS_CONTINUE;
	}
	if (TextInsensitiveCompare(name, X_SPAM_REPORT) == 0) {
		data->hasSpamReport = 1;
		return SMFIS_CONTINUE;
	}
	if (TextInsensitiveCompare(name, X_SPAM_STATUS) == 0) {
		data->hasSpamStatus = 1;
		return SMFIS_CONTINUE;
	}
	
	if (TextInsensitiveCompare(name, "Precedence") == 0) {
		for (i = 0; i <= PRECEDENCE_SPECIAL_DELIVERY; i++) {
			if (TextInsensitiveCompare(value, precedence[i]) == 0) {
				data->precedence = i;
				break;
			}
		}
		return SMFIS_CONTINUE;
	}

	(void) BufAddString(data->headers, name);
	(void) BufAddBytes(data->headers, ": ", 0, 2);
	(void) BufAddString(data->headers, value);
	(void) BufAddBytes(data->headers, "\r\n", 0, 2);

	return SMFIS_CONTINUE;
}

sfsistat
filterEndHeaders(SMFICTX *ctx)
{
	workspace data;

	if ((data = (workspace) smfi_getpriv(ctx)) == (workspace) 0)
		return nullWorkspaceError(ctx);

	if (debug & DEBUG_TRACE)
		syslog(LOG_DEBUG, "%s: filterEndHeaders(%lx)", data->qid, (long) ctx);

	if (data->byPass)
		return SMFIS_CONTINUE;

	(void) BufAddBytes(data->headers, "\r\n", 0, 2);

	return SMFIS_CONTINUE;
}

sfsistat
filterBody(SMFICTX *ctx, unsigned char *chunk, size_t size)
{
	long length, n;
	workspace data;
	char *msg_size, *if_addr, *client_addr, *client_name, *block;

	if ((data = (workspace) smfi_getpriv(ctx)) == (workspace) 0)
		return nullWorkspaceError(ctx);

	if (debug & DEBUG_TRACE)
		syslog(LOG_DEBUG, "%s: filterBody(%lx, '%.20s...', %lu)", data->qid, (long) ctx, chunk, size);

	if (data->byPass || data->isSent)
		return SMFIS_CONTINUE;

	if (smfi_getsymval(ctx, "{if_addr}") == (char *) 0) {
		syslog(LOG_INFO, "%s: loopback address, skipping scan", data->qid);
		return SMFIS_CONTINUE;
	}

 	/* Only process the headers and first body chunk. */
 	data->isSent = 1;

	/* Insert a simulated Received: header for this server into the
	 * header block being sent to spamd. It appears that this header
	 * is never given to the milter and without it, some tests in
	 * SpamAssassin concerning Received headers are skewed.
	 *
	 * This simulated header will be a stripped down version of
	 * what Sendmail would insert.
	 */
	client_addr = smfi_getsymval(data->ctx, "{client_addr}");
	client_name = smfi_getsymval(data->ctx, "{client_name}");

	if (client_addr == (char *) 0 || client_name == (char *) 0) {
		syslog(LOG_ERR, "%s: {client_addr} and {client_name} must be available, see cf/README `confMILTER_MACROS_CONNECT'", data->qid);
	} else {
		length = snprintf(
			data->line, SMTP_LINE_SIZE, "Received: from %s (%s [%s]) by %s id %s; ", 
			data->helo, client_name, client_addr, hostname, data->qid
		);			

		if (0 < (n = TimeStampAdd(data->line + length, SMTP_LINE_SIZE - length - 2))) {
			data->line[length + n    ] = '\r';
			data->line[length + n + 1] = '\n';

			if (data->indexReceived < 0)
				data->indexReceived = 0;

			(void) BufInsertBytes(data->headers, data->indexReceived, data->line, 0, length);

			if (debug & DEBUG_SPAMD)
				syslog(LOG_DEBUG, "%s: after length=%ld headers {%s}", data->qid, BufLength(data->headers), BufBytes(data->headers));		
		}
	}
	
	data->server = SocketOpenTcpClientWait(spamdHost, spamdPort, socketTimeout);

	if (data->server == (Socket) 0) {
		syslog(LOG_ERR, "%s: failed to connect to spamd %s,%d", data->qid, spamdHost, spamdPort);
		goto error0;
	}

	if (debug & DEBUG_TRACE)
		syslog(LOG_DEBUG, "%s: SPAMD Content-Length=%ld", data->qid, BufLength(data->headers) + size);

	/* Ask for a CHECK or REPORT only and we'll handle adding the X-Spam-* headers. */
	length = snprintf(
		data->line, SMTP_LINE_SIZE, "%s SPAMC/1.2\r\nContent-Length: %ld\r\n",
		spamcCommand, BufLength(data->headers) + size
	);
	if (!writeline(data, data->line, length))
		goto error1;

	/* If the spamd user is defined, then we'll use it if forced or
	 * multiple recipients given. Otherwise fall back on the last
	 * defined local recipient if any.
	 */
	if (spamdUser != (char *) 0 && (forceSpamdUser || 1 < VectorLength(data->rcpts))) {
		/* Free NULL or a previous local user. */
		free(data->localRcpt);
		data->localRcpt = (char *) 0;
	}

	/* When no local recipient defined, then fallback on the default
	 * spamd user, which might be undefined.
	 */
	if (data->localRcpt == (char *) 0)
		data->localRcpt = spamdUser;

	if (data->localRcpt != (char *) 0) {
		if (debug & DEBUG_TRACE)
			syslog(LOG_DEBUG, "%s: SPAMD User=%s", data->qid, data->localRcpt);
		length = snprintf(data->line, SMTP_LINE_SIZE, "User: %s\r\n", data->localRcpt);
		if (!writeline(data, data->line, length))
			goto error1;
	}

	if (!writeline(data, "\r\n", 2))
		goto error1;

	if (!writechunk(data, BufBytes(data->headers), BufLength(data->headers)))
		goto error1;

	if (!writechunk(data, chunk, size))
		goto error1;

	/* Signal EOF to SPAMD so that it can begin processing. */
	SocketShutdownNow(data->server, SHUT_WR);

	return SMFIS_CONTINUE;
error1:
	SocketClose(data->server);
	data->server = (Socket) 0;
error0:
	return SMFIS_CONTINUE;
}

static int
headerAddOrChange(SMFICTX *ctx, char *field, char *value, int index, int present)
{
	if (present)
		return smfi_chgheader(ctx, field, index, value);
		
	return smfi_addheader(ctx, field, value);
}

sfsistat
filterEndMessage(SMFICTX *ctx)
{
	workspace data;
	ParsePath *path;
	float score, threshold;
	int i, isSpam, length, iscore;

	if ((data = (workspace) smfi_getpriv(ctx)) == (workspace) 0)
		return nullWorkspaceError(ctx);

	if (debug & DEBUG_TRACE)
		syslog(LOG_DEBUG, "%s: filterEndMessage(%lx)", data->qid, (long) ctx);

	/* Send a copy of all mail to the archive address. */
	if (copyAllTo != (char *) 0 && smfi_addrcpt(ctx, copyAllTo) == MI_FAILURE)
		syslog(LOG_ERR, "%s: copy-all-to <%s> failed", data->qid, copyAllTo);

	/* Skip the remainder of this function if there was an
	 * earlier error or if we did not start a spamd session.
	 */
	if (data->byPass || data->server == (Socket) 0)
		return SMFIS_CONTINUE;

	/* Fetch spamd result and add it to the headers. */
	if (readline(data, data->line, SMTP_LINE_SIZE) <= 0) {
		syslog(LOG_ERR, "%s: SPAMD status line failure", data->qid);
		return SMFIS_CONTINUE;
	}

	if (strncmp(data->line, "SPAMD/1.1 0 EX_OK", sizeof ("SPAMD/1.1 0 EX_OK")) != 0) {
		syslog(LOG_ERR, "%s: SPAMD failure: %s", data->qid, data->line);
		return SMFIS_CONTINUE;
	}

	/* Set to -2 for undefined, since TextInsensitiveStartsWith() 
	 * returns -1 on no match or a zero/postive length for a match.
	 */
	isSpam = -2;
	
	while (0 < readline(data, data->line, SMTP_LINE_SIZE)) {
		if (0 < TextInsensitiveStartsWith(data->line, "Spam:")) {
			/* And the verdict is? */
			isSpam = 0 < TextInsensitiveStartsWith(data->line, "Spam: True");
			(void) headerAddOrChange(ctx, X_SPAM_FLAG, isSpam ? "YES" : "NO", 1, data->hasSpamFlag);

			/* Extract the score. */
			if (sscanf(data->line, "%*[^;]; %f / %f", &score, &threshold) != 2) {
				syslog(LOG_ERR, "%s: Spam: header parse error: %s", data->qid, data->line);
				return SMFIS_CONTINUE;
			}
		}
		
		/* Ignore any other headers. */
	}

	length = snprintf(data->line, SMTP_LINE_SIZE, MILTER_VERSION " (%s [%s]); pass=%s; ", hostname, hostaddr, isSpam == -2 ? "UNKNOWN" : (isSpam ? "NO" : "YES"));			
	length += TimeStampAdd(data->line + length, SMTP_LINE_SIZE - length);
#ifdef DONT_DO_THIS
/* If you add a CRLF to the end of a header, sendmail reports a 
 * POSSIBLE ATTACK from... log message when it goes to append
 * its newline.
 */
	(void) strncpy(data->line + length, "\r\n", SMTP_LINE_SIZE - length);
#endif
	(void) smfi_addheader(ctx, X_SCANNED_BY, data->line);
	
	if (isSpam == -2) {
		syslog(LOG_ERR, "%s: missing Spam: result header", data->qid);
		return SMFIS_CONTINUE;
	}

	/* Add a pretty header. I think this one is useless. */
	(void) sprintf(data->line, "%s, hits=%.2f required=%.2f", isSpam ? "YES" : "NO", score, threshold);
	(void) headerAddOrChange(ctx, X_SPAM_STATUS, data->line, 1, data->hasSpamStatus);

	/* Add another header that can be parsed with by a regular expression.
	 * This one is useful for filtering different levels of spam. Note that
	 * we initially use a non-regular-expression character for writing the
	 * level.
	 */
	if (SMTP_LINE_SIZE <= score)
		/* X-Spam-Level: xxx...x\r\n\0 */
		score = SMTP_LINE_SIZE - sizeof (X_SPAM_LEVEL) - 5;

	iscore = (int) score;
	for (i = 0; i < iscore; i++)
		data->line[i] = DEFAULT_SPAM_LEVEL_CHARACTER;
	data->line[i] = '\0';
	
	(void) headerAddOrChange(ctx, X_SPAM_LEVEL, data->line, 1, data->hasSpamLevel);

	/* REUSE this buffer to collect the report. */
	BufSetLength(data->headers, 0);

	/* SpamAssassin 2.5x prefixed a redundant score line for a report. */
	length = readline(data, data->line, SMTP_LINE_SIZE);
	if (strspn(data->line, "1234567890-./") != length) {
		/* Remove trailing whitespace. */
		while (0 < length-- && isspace(data->line[length]))
			data->line[length] = '\0';

		/* Only add non-empty lines. */
		if (data->line[0] != '\0') {
			BufAddString(data->headers, data->line);
			BufAddString(data->headers, "\n  ");
		}
	}

	/* Collect report lines until read error or EOF. */
	while (0 <= (length = readline(data, data->line, SMTP_LINE_SIZE))) {
		/* Remove trailing whitespace. */
		while (0 < length-- && isspace(data->line[length]))
			data->line[length] = '\0';

		if (length <= 0) {
			/* Replace a empty line with something that looks blank. */
			BufAddString(data->headers, "____\n  ");
		} else {
			BufAddString(data->headers, data->line);
			BufAddString(data->headers, "\n  ");
		}
	}

	/* Done with spamd session. */
	SocketClose(data->server);
	data->server = (Socket) 0;

	/* Remove our trailing whitespace "\r\n  ". */
	BufSetLength(data->headers, BufLength(data->headers)-3);

	if (alwaysAddReport)
		(void) headerAddOrChange(ctx, X_SPAM_REPORT, BufBytes(data->headers), 1, data->hasSpamReport);

	if (isSpam) {
		if (!alwaysAddReport)
			(void) headerAddOrChange(ctx, X_SPAM_REPORT, BufBytes(data->headers), 1, data->hasSpamReport);

		/* Tag the subject if not already. */
		if (TextInsensitiveStartsWith(data->subject, subjectTag) < 0) {
			if (appendScoreToSubjectTag)			
				(void) snprintf(data->line, sizeof data->line, "%s (%.2f/%.2f) %s", subjectTag, score, threshold, data->subject);
			else
				(void) snprintf(data->line, sizeof data->line, "%s %s", subjectTag, data->subject);

			(void) headerAddOrChange(ctx, "Subject", data->line, 1, data->hasSubject);
		}

		if (copySpamTo != (char *) 0) {
			/* The -F option can only redirect to individual mail folders when
			 * the entire recipient list is local. Otherwise send to the system
			 * spam box for processing.
			 *
			 * The reason for this restriction is that +detail information is 
			 * site specific and so adding this information into outbound mail
			 * is useless. Also when the mail is a mix of local and outbound mail
			 * (for example local user sending to multiple addresses locally and
			 * off-site), then the question is what to do with the outbound mail
			 * addresses: 
			 *
			 * a) drop addresses that can't be transformed; not ideal since some
			 * will receive the message in a spam folder and others won't recieve
			 * it at all.
			 *
			 * b) forward to the system collection box; this is more in line with
			 * previous behaviour of the milter and allows the postmaster to 
			 * retrain or whitelist within SpamAssassin.
			 *
			 * c) reject the whole message outright; this is a nice choice, since
			 * the sender gets immediate feedback that there is a problem, but 
			 * not so easy to retrain SpamAssassin.
			 *
			 * d) don't scan outbound mail; this involves determining whether the
			 * sender is truly a local user (login account, aliases database, 
			 * virtual user box), which increases the code complexity. See -f option.
			 */
			int sendToSystemSpamBox = redirectSpam != REDIRECT_TO_FORMAT || VectorLength(data->rcpts) != data->localRcptCount;
		
			if (redirectSpam) {
				/* Remove previous list of recipients from Sendmail's list
				 * and record them in the redirected message in case it was
				 * not spam and needs to be resent later. Sometimes the To:
				 * or Cc: headers are useless ie. undisclosed recipients.
				 */
				for (i = 0; i < VectorLength(data->rcpts); i++) {
					if ((path = VectorGet(data->rcpts, i)) == (ParsePath *) 0)
						continue;

					if (smfi_addheader(ctx, X_ORIGINAL_RECIPIENT, path->address.string) == MI_SUCCESS)
						(void) smfi_delrcpt(ctx, path->address.string);
				}

				/* If -F was given, then modify the recipient list so that
				 * each local recipient is modified according to a sepcified
				 * format.
				 */
				if (!sendToSystemSpamBox) {
					for (i = 0; i < VectorLength(data->rcpts); i++) {
						if ((path = VectorGet(data->rcpts, i)) == (ParsePath *) 0)
							continue;
							
						/* We cannot modify non-local recipients, because
						 * plus details are a sendmail thang and the remote
						 * end might handle spam differently.
						 */
						if (!path->isLocal)
							continue;
						
						(void) formatPath(data->line, sizeof(data->line), rcptReformat, path);
					
						if (debug & DEBUG_TRACE)
							syslog(LOG_DEBUG, "%s: modified recipient=%s", data->qid, data->line);
						
						if (smfi_addrcpt(ctx, data->line) == MI_FAILURE)
							syslog(LOG_ERR, "%s: redirect-spam-to %s failed ", data->qid, data->line);
					}
				}
			}

			if (sendToSystemSpamBox && smfi_addrcpt(ctx, copySpamTo) == MI_FAILURE) {
				syslog(LOG_ERR, "%s: copy-spam-to <%s> failed ", data->qid, copySpamTo);
				goto error1;
			}
		}
	} else {
		/* Send a copy of clean mail to an archive address. */
		if (copyHamTo != (char *) 0 && smfi_addrcpt(ctx, copyHamTo) == MI_FAILURE)
			syslog(LOG_ERR, "%s: copy-ham-to <%s> failed ", data->qid, copyHamTo);
	}

error1:
	/* REUSE this buffer to build RCPT list string. */
	BufSetLength(data->headers, 0);

	/* Save copy of recipient list into a string for log information. */
	for (i = 0; i < VectorLength(data->rcpts); i++) {
		if ((path = VectorGet(data->rcpts, i)) != (ParsePath *) 0) {
			BufAddString(data->headers, path->address.string);
			BufAddString(data->headers, ">,<");
		}
	}
	
	/* Remove trailing ",<". */
	BufSetLength(data->headers, BufLength(data->headers)-2);

	syslog(LOG_INFO, "%s: spam=%s score=%.2f required=%.2f subject='%s' mail=<%s> rcpts=<%s", data->qid, isSpam ? "YES" : "NO", score, threshold, data->subject, data->mail->address.string, BufBytes(data->headers));		
error0:
	if (isSpam) {
		(void) setReply(data, 550, "5.7.1", "message was identified as junk mail");

		if (threshold + rejectExcessScore <= score)
			return PRECEDENCE_FIRST_CLASS <= data->precedence ? SMFIS_REJECT : SMFIS_DISCARD;
		if (threshold + discardExcessScore <= score)
			return SMFIS_DISCARD;
#ifdef SMFIS_REJECT_AND_DELIVER
		if (threshold + rejectAndDeliverExcessScore <= score)
			return SMFIS_REJECT_AND_DELIVER;
#endif
	}

	return SMFIS_CONTINUE;
}

/*
 * Close and release per-connection resources.
 */
sfsistat
filterClose(SMFICTX *ctx)
{
	char *qid;
	workspace data = (workspace) smfi_getpriv(ctx);

	if (data != (workspace) 0) {
		qid = data->qid;
		
		/* We have to do this, otherwise we get a warning. */
		(void) smfi_setpriv(ctx, (void *) 0);

		if (data->localRcpt != spamdUser)
			free(data->localRcpt);

		VectorDestroy(data->rcpts);
		BufDestroy(data->headers);
		SocketClose(data->server);
		free(data->mail);
		free(data);

		if (debug & DEBUG_TRACE)
			syslog(LOG_DEBUG, "%s: destroyed per-connection resources", qid);
	}

	return SMFIS_CONTINUE;
}

/***********************************************************************
 ***  Milter Definition Block
 ***********************************************************************/

struct smfiDesc smfilter = {
	MILTER_NAME,		/* filter name */
	SMFI_VERSION,		/* version code -- do not change */
	SMFIF_ADDHDRS|SMFIF_CHGHDRS|SMFIF_ADDRCPT|SMFIF_DELRCPT,	/* flags */
	filterOpen,		/* connection info filter */
	filterHelo,		/* SMTP HELO command filter */
	filterMail,		/* envelope sender filter */
	filterRcpt,		/* envelope recipient filter */
	filterHeader,		/* header filter */
	filterEndHeaders,	/* end of header */
	filterBody,		/* body block filter */
	filterEndMessage,	/* end of message */
	NULL,			/* message aborted */
	filterClose,		/* connection cleanup */
};

/***********************************************************************
 *** Startup
 ***********************************************************************/

void
signalExit(int signum)
{
	(void) signal(signum, SIG_IGN);
	syslog(LOG_ERR, "signal %d received, program exit", signum);
	exit(0);
}

void
atExitCleanup()
{
# ifdef HAVE_BERKELEY_DB
	if (smdbAccess != (smdb *) 0 && smdbAccess->db != (DB *) 0)
#  if DB_VERSION_MAJOR == 1
		smdbAccess->db->close(smdbAccess->db);
#  else
		smdbAccess->db->close(smdbAccess->db, 0);
#  endif
	VectorDestroy(smdbAliases);
	MutexDestroy(theMutex);
# endif
	/* Unlink the named socket ourselves since it appears that
	 * libmilter doesn't do this. See Sendmail 8.13.0 change log.	 
	 */
	if (strncmp(milterSocket, "unix:", 5) == 0) 
		(void) unlink(milterSocket + 5);
	else if (strncmp(milterSocket, "local:", 6) == 0)
		(void) unlink(milterSocket + 6);
}

int
main(int argc, char **argv)
{
	int ch, rc;
	struct stat sb;
	struct group *gr;
	struct passwd *pw;
	struct hostent *host;

	openlog(MILTER_NAME, LOG_PID, LOG_MAIL);

	if (signal(SIGQUIT, signalExit) == SIG_ERR) {
		(void) fprintf(stderr, "failed to set SIGQUIT handler: %s (%d)\n", strerror(errno), errno);
		syslog(LOG_ERR, "failed to set SIGQUIT handler: %s (%d)", strerror(errno), errno);
		return 1;
	}

	syslog(LOG_DEBUG, "process uid=%d gid=%d before", getuid(), getgid());

	if ((pw = getpwnam(RUN_AS_USER)) == (struct passwd *) 0) {
		(void) fprintf(stderr, "user \"%s\" not found\n", RUN_AS_USER);
		syslog(LOG_ERR, "user \"%s\" not found", RUN_AS_USER);
		return 1;
	}

	if ((gr = getgrnam(RUN_AS_GROUP)) == (struct group *) 0) {
		(void) fprintf(stderr, "group \"%s\" not found\n", RUN_AS_GROUP);
		syslog(LOG_ERR, "group \"%s\" not found", RUN_AS_GROUP);
		return 1;
	}

	/* Avoid running as root user and/or group. */
	if (getuid() == 0 && pw->pw_uid != 0 && gr->gr_gid != 0) {
		(void) setgid(gr->gr_gid);
		(void) setuid(pw->pw_uid);

		syslog(LOG_DEBUG, "process uid=%d gid=%d after", getuid(), getgid());
	}

	debug = DEBUG_INFO;
	alwaysAddReport = 0;
	copyAllTo = (char *) 0;
	copyHamTo = (char *) 0;
	copySpamTo = (char *) 0;
	socketTimeout = DEFAULT_SOCKET_TIMEOUT;
	subjectTag = DEFAULT_SPAM_SUBJECT_TAG;
	appendScoreToSubjectTag = 0;
	spamdHost = DEFAULT_SPAMD_HOST;
	spamdPort = DEFAULT_SPAMD_PORT;
	spamdUser = (char *) 0;
	forceSpamdUser = 0;
	redirectSpam = COPY_TO;
	untrustedRcptLimit = 0;
	spamcCommand = DEFAULT_SPAMC_COMMAND;
	rejectExcessScore = ~(unsigned long)0 >> 1;
	discardExcessScore = ~(unsigned long)0 >> 1;
#ifdef SMFIS_REJECT_AND_DELIVER
	rejectAndDeliverExcessScore = ~(unsigned long)0 >> 1;
#endif

	while ((ch = getopt(argc, argv, getoptString)) != -1) {
		switch (ch) {
		case 'A':
			alwaysAddReport = 1;
			spamcCommand = "REPORT";
			break;
		case 'C':
			alwaysAddReport = 0;
			spamcCommand = "CHECK";
			break;
		case 'a':
			copyAllTo = optarg;
			break;
		case 'b':
			copySpamTo = optarg;
			redirectSpam = COPY_TO;
			break;
		case 'B':
			copySpamTo = optarg;
			redirectSpam = REDIRECT_TO;
			break;
		case 'F':
			rcptReformat = optarg;
			redirectSpam = REDIRECT_TO_FORMAT;
			break;
		case 'c':
			copyHamTo = optarg;
			break;
		case 'd':
			discardExcessScore = strtol(optarg, (char **) 0, 10);
			if (discardExcessScore < 0) {
				(void) fprintf(stderr, "-d value must be positive");
				(void) fprintf(stderr, usageMessage);
			}
			break;
		case 'f':
			sendmailCf = optarg;
			break;
		case 'l':
			untrustedRcptLimit = strtol(optarg, (char **) 0, 10);
			if (untrustedRcptLimit < 0) {
				(void) fprintf(stderr, "-l value must be positive");
				(void) fprintf(stderr, usageMessage);
			}
			break;
		case 't':
			socketTimeout = strtol(optarg, (char **) 0, 10);
			break;
		case 'r':
			rejectExcessScore = strtol(optarg, (char **) 0, 10);
			if (rejectExcessScore < 0) {
				(void) fprintf(stderr, "-r value must be positive");
				(void) fprintf(stderr, usageMessage);
			}
			break;
#ifdef SMFIS_REJECT_AND_DELIVER
		case 'R':
			rejectAndDeliverExcessScore = strtol(optarg, (char **) 0, 10);
			if (rejectAndDeliverExcessScore < 0) {
				(void) fprintf(stderr, "-R value must be positive");
				(void) fprintf(stderr, usageMessage);
			}
			break;
#endif
		case 's':
			subjectTag = optarg;
			break;
		case 'S':
			appendScoreToSubjectTag = 1;
			break;
		case 'u':
			spamdUser = optarg;
			forceSpamdUser = 0;
			break;
		case 'U':
			spamdUser = optarg;
			forceSpamdUser = 1;
			break;
		case 'v':
			debug = strtol(optarg, (char **) 0, 10);
			break;
		case 'H':
			spamdHost = optarg;
			break;
		case 'P':
			spamdPort = (int) strtol(optarg, (char **) 0, 10);
			break;
		default:
			syslog(LOG_ERR, "unknown command line option -%c", optopt);
			/*@fallthrough@*/
		case 'h':
		case '?':
			(void) fprintf(stderr, usageMessage);
			return 2;
		}
	}

	socketTimeout *= 1000;

	if (optind + 1 != argc) {
		syslog(LOG_ERR, "missing command line argument");
		(void) fprintf(stderr, usageMessage);
		return 2;
	}

	milterSocket = argv[optind];
	umask(S_IXUSR | S_IRWXG | S_IRWXO);

	if (atexit(atExitCleanup)) {
		syslog(LOG_ERR, "atexit() failed\n");
		return 1;
	}

#if defined(HAVE_BERKELEY_DB)
	if (stat(THE_MUTEX, &sb) == 0) {
		(void) fprintf(stderr, MILTER_NAME " already running or previously leaked semaphore, see " THE_MUTEX "\n");
		syslog(LOG_ERR, MILTER_NAME " already running or previously leaked semaphore, see " THE_MUTEX);
		return 1;
	}
	
	if ((theMutex = MutexCreate(THE_MUTEX)) == (void *) 0) {
		syslog(LOG_ERR, "failed to allocate mutex\n");
		return 1;
	}

	if (MutexPermission(theMutex, 0600, pw->pw_uid, gr->gr_gid) < 0) {
		syslog(LOG_ERR, "failed to set mutex permissions\n");
		return 1;
	}
#endif

	/* Specifying the -s option enable access and aliases database support. */
	if (sendmailCf != (char *) 0) {
		(void) readSendmailCf(sendmailCf, SMDB_OPEN_ACCESS);
#ifdef HAVE_BERKELEY_DB
		if (smdbAccess->db == (DB *) 0)
                        syslog(LOG_WARNING, "access database support disabled");
#endif
	}

	if (debug & DEBUG_DATABASE)
		smdbSetDebugMask(SMDB_DEBUG_ALL);

	if (smfi_setconn(milterSocket) == MI_FAILURE) {
		syslog(LOG_ERR, "smfi_setconn() failed\n");
		return 1;
	}

	if (strncmp(milterSocket, "unix:", 5) == 0) {
		unlink(milterSocket + 5);
	} else if (strncmp(milterSocket, "local:", 6) == 0) {
		unlink(milterSocket + 6);
	}

	if (smfi_register(smfilter) == MI_FAILURE) {
		syslog(LOG_ERR, "smfi_register() failed\n");
		return 1;
	}

	if (debug & DEBUG_SOCKETS)
		SocketSetDebugMask(SOCKET_DEBUG_ALL);

	if (SocketInit()) {
		syslog(LOG_ERR, "SocketInit() error\n");
		return 1;
	}

	gethostname(hostname, sizeof hostname);
	if (debug & DEBUG_TRACE)
		syslog(LOG_DEBUG, "gethostname() hostname=%s", hostname);

#if defined(HAVE_GETHOSTBYNAME2)
	/* IPv4 and IPv6 solution */
	if ((host = gethostbyname2(hostname, AF_INET)) == (struct hostent *) 0
	&&  (host = gethostbyname2(hostname, AF_INET6)) == (struct hostent *) 0) {
#elif defined(HAVE_GETHOSTBYNAME)
	/* IPv4 only solution */
	if ((host = gethostbyname(hostname)) == (struct hostent *) 0) {
#else
# error "require gethostbyname() or gethostbyname2()"
#endif
		syslog(LOG_ERR, "gethostbyname(%s) failed: %s (%d)", hostname, strerror(h_errno), h_errno);
		return 1;
	}

	if (inet_ntop(host->h_addrtype, host->h_addr_list[0], hostaddr, sizeof hostaddr) == NULL) {
		syslog(LOG_ERR, "inet_ntop() failed");
		return 1;
	}

#ifdef CLOSE_STANDARD_HANDLES
	/* Don't need these and they just use up resources. */
	fclose(stderr);
	fclose(stdout);
	fclose(stdin);
#endif

	/* Display version information in log. */
	syslog(LOG_INFO, MILTER_VERSION " " MILTER_COPYRIGHT);
	syslog(LOG_INFO, LIBSNERT_VERSION " " LIBSNERT_COPYRIGHT);

#if defined(HAVE_BERKELEY_DB)
# if DB_VERSION_MAJOR > 1
	syslog(LOG_INFO, "%s", db_version((int *) 0, (int *) 0,(int *) 0));
# else
	syslog(LOG_INFO, "Built with Berkeley DB");
# endif
#else 
	syslog(LOG_INFO, "Built without Berkeley DB support");
#endif

	return smfi_main();
}
